/*-----------------------------------------------------*/
/*--	Micro-EMPIX CORE  (Micro.c)                   -*/
/*--    Ver 1.0 (c) 2004                              -*/
/*--	E             -*/
/*--    ==================================            -*/
/*--    .                  -*/
/*--    .                        -*/
/*--    . .                      -*/
/*-----------------------------------------------------*/

#include "micro.h"

extern void P1();
extern void P2();
extern void Initializer();
/* ====================================================*/
/* ==========   ============*/
/* ====================================================*/

struct pcb
{
	 char  pstate;      		/*     */
	 int pid;		   			/*  PID   */
	 int pprio;         		/*    */
	 int pregs[PNREGS]; 		/*     */
	 char pname[PNMLEN];		/*     */
	 int type;	       			/*    */
} proctab[MAXPROC];


KERNEL_INT current; /*    */
KERNEL_INT numproc; /*      .. */
KERNEL_INT stacks[MAXPROC][STACKSIZE]; /*     */
KERNEL_INT preempt; /*        */
KERNEL_INT preregs[8]; /*      */

/* ====================================================*/
/* =========    ==========*/
/* ====================================================*/

KERNEL_INT semaphores [MAXSEM][MAXPROC]; /*   */
KERNEL_INT semavalue [MAXSEM]; /*      */


static void interrupt (*IrqHandler_For_Clk)(); /*    */

/* ====================================================*/
/* ==========    / ===========*/
/* ====================================================*/
static unsigned char SCR_Current_Color; 	/*       */
											/*      */
											/*   BIOS software interrupts */

#define SIRQS_ISOLATED 1

/*        */
/*    Data segment       */
/*   interrupt handler,       */
/*  global variables.  data segment    */
/*   interrupt handler   28 (DOS IDLE LOOP).  */
/*          interrupt handler */
/* (OldInt28) */
#define ds_store (4*0x28) /* In place of DOS IDLE loop interrupt */
static unsigned int OldInt28;

#define IRQVEC_FOR_SER_P1 0xc
#define IRQVEC_FOR_SER_P2 0xb

static int SER_baseport1;
static int SER_baseport2;

/* For Serial Port Hardware Flow Control */
static unsigned char HardwareFlowControl1;
static unsigned char HardwareFlowControl2;
#define MAXCHARS_BEFORE_STOPPING 6
static unsigned int CharsSent1;
static unsigned int CharsSent2;
static unsigned char PrevSignal1;
static unsigned char PrevSignal2;

/* For Interrupt Driven Serial Ports */
/*   interrupt handler  COM-1,      */
static void interrupt (*IrqHandler_For_Com1)();
static void interrupt (*IrqHandler_For_Com2)();
static unsigned char IrqHandlerForCom1_Changed;
static unsigned char IrqHandlerForCom2_Changed;

static void (*SER1_func)(void);
static void (*SER2_func)(void);

/* Variables used in simplified versions of Serial Port drivers */
#define BUFLEN 1024
#define ISEM1 (MAXSEM - 1)
#define OSEM1 (MAXSEM - 2)
#define ISEM2 (MAXSEM - 3)
#define OSEM2 (MAXSEM - 4)

struct QSR_ERROR QSR_Error;

/*  buffer         */
static char SER1_InBuf[BUFLEN];
static char *SER1_InBufPtrRd, *SER1_InBufPtrWr;

/*  buffer         */
static char SER2_InBuf[BUFLEN];
static char *SER2_InBufPtrRd, *SER2_InBufPtrWr;

/*==========================================================================================*/
/*=====================    ============================================*/
/*==========================================================================================*/


static unsigned char port_in(int baseport, int offset)
{
        unsigned char mret = 0;

        asm MOV DX,baseport
        asm ADD DX,offset
        asm IN AL,DX
        asm MOV BYTE PTR mret,AL

        return mret;
}

static void port_out(int baseport, int offset, unsigned char val)
{
        asm MOV DX,baseport
        asm ADD DX,offset
        asm MOV AL,val
        asm OUT DX,AL
}

/*   Cbit = 0,  LCR */

static void UnSetControlBit(int baseport)
{
        asm MOV DX,baseport
        asm ADD DX,3
        asm IN AL,DX
        asm AND AL,7FH
        asm OUT DX,AL
}

/*   Cbit = 1,  LCR */

static void SetControlBit(int baseport)
{
        asm MOV DX,baseport
        asm ADD DX,3
        asm IN AL,DX
        asm OR AL,80H
        asm OUT DX,AL
}

/*   Baud Rate */

static SetBaudRate(int baseport, BAUD_RATE brate)
{
    /* Setting BaudRate */
    /* Set line control 7th bit to 1, so that we can talk to DLL and DLM */
    SetControlBit(baseport);
	/* Computing Baud rate for DLL */
	asm MOV BX,brate
	asm AND BX,0FFH
    /* Talk to DLL (Baud rate divisor least significant byte) */
	asm MOV DX,baseport;
	asm MOV AL,BL
	asm OUT DX,AL
	/* Computing Baud rate for DLM */
	asm MOV BX,brate
	asm AND BX,0FF00H
	/* Talk to DLM (Baud rate divisor most significant byte) */
    asm MOV DX,baseport;
	asm INC DX
	asm MOV AL,BH
	asm OUT DX,AL
    /* Make 7th bit of LCR is 0 */
    UnSetControlBit(baseport);
}

/*   LCR, o    (8 bits), 1 stop bit    */

static void SetWordLength(int baseport)
{
	/* Setting Word length, Stop bits and Parity */
    port_out(baseport, 3, 0x3); /* 8 bits word length, 1 stop bit and no parity */
}


/*   LSR, MSR & IIR     interrupts */

void ClearLSR_MSR_IIR(int baseport)
{
        /* Clearing LSR, MSR, IIR */
        asm MOV DX,baseport
        asm ADD DX,5
        asm IN AL,DX
        asm INC DX
        asm IN AL,DX
        asm SUB DX,4
        asm IN AL,DX
}

/*         16550,  */
/*  FIFO buffers        */
/* (    ) */

static void ClearAndEnableFIFO(int baseport, unsigned char chh)
{
        /* Clear internal 16550 buffers and enable FIFO buffers */
        /* Enable new buffers */
        port_out(baseport, 2, chh | 3);
}

/*        interrupts   16550,   */
/*          16550,     */
/* interrupt.     1, 4, 8  12.      FIFO Control register */

static void NrOfCharsBefReadIRQ(int baseport, unsigned char chh)
{
        /* Set number of chars before each interrupt */
        port_out(baseport, 2, chh | 3);
}

/*          interrupt   UART,  IER */

static void IER_Enable(int baseport, unsigned char IR_Byte)
{
        /* Insane measure. Make sure that 7th bit of LCR is 0 */
        UnSetControlBit(baseport);
        /* Enable interrupts in IER */
        port_out(baseport, 1, IR_Byte);
}

/*     MCR */

static void MCR_Enable(int baseport)
{
        /* Enable interrupts in Modem control register */
        port_out(baseport, 4, 0x0b);
}

/*        interrupts  IER */

static void IER_Disable(int baseport)
{
        /* Disable interrupts in IER */
        port_out(baseport, 1, 0);
}

/*          ,   interrupt hanlder */
/*    .              SER_func, */
/*                struct actions */
/*     SER_InitializePort */

static void interrupt InterruptHandlerForCOM1()
{
	asm PUSH AX
	asm PUSH BX
	asm PUSH CX
	asm PUSH DX
	asm PUSH SI
	asm PUSH DI
	asm PUSH DS
	asm PUSH ES

    /* Restore the original DS */
	asm XOR AX,AX
	asm MOV ES,AX
	asm MOV BX,ds_store
	asm MOV DS,WORD PTR ES:[BX]

	asm XOR AX,AX
	asm PUSH AX

	SER1_func();

	asm ADD SP,2

	/* Tell the 8259 PIC chip that we have served the interrupt */
	asm MOV AL,20H
	asm OUT 20H,AL

	asm POP ES /*   */
	asm POP DS
	asm POP DI
	asm POP SI
	asm POP DX
	asm POP CX
	asm POP BX
	asm POP AX
}

/*      */

static void interrupt InterruptHandlerForCOM2()
{
	asm PUSH AX
	asm PUSH BX
	asm PUSH CX
	asm PUSH DX
	asm PUSH SI
	asm PUSH DI
	asm PUSH DS
	asm PUSH ES

    /* Restore the original DS */
	asm XOR AX,AX
	asm MOV ES,AX
	asm MOV BX,ds_store
	asm MOV DS,WORD PTR ES:[BX]

	asm XOR AX,AX
	asm PUSH AX

        SER2_func();

	asm ADD SP,2

	/* Tell the 8259 PIC chip that we have served the interrupt */
	asm MOV AL,20H
	asm OUT 20H,AL

	asm POP ES
	asm POP DS
	asm POP DI
	asm POP SI
	asm POP DX
	asm POP CX
	asm POP BX
	asm POP AX
}

/*        portnum.       */
/*      SER.  brate   Baud Rate (   enum),*/
/*  HWDFlowControl     0,    actions   NULL,    */
/*     polling.          interrupts */
/*            actions */

KERNEL_VOID SER_InitializePort(SER_PORT portnum, BAUD_RATE brate, const unsigned char HWDFlowControl, struct IrqConf *actions)
{
	int baseport;
	unsigned char chh;

	if(portnum != SER_P1 && portnum != SER_P2) return;
    if(portnum == SER_P1 && SER_baseport1) /* Serial already initialized */
		return;
    if(portnum == SER_P2 && SER_baseport2) /* Serial already initialized */
        return;
    asm PUSHF
    asm CLI

    /* The following must be done only the first time */
    if(!SER_baseport1 && !SER_baseport2)
    {
		/* Save the contents of the interrupt handler */
        asm XOR AX,AX
        asm MOV ES,AX
        asm MOV BX,ds_store
        asm MOV AX, WORD PTR ES:[BX]
        asm MOV WORD PTR OldInt28, AX
        /* Insert our current DS */
        asm XOR AX,AX
        asm MOV ES,AX
        asm MOV AX,DS
        asm MOV BX,ds_store
        asm MOV WORD PTR ES:[BX], AX
    }

    /* Get Port Num from Bios */
    asm MOV BX,40H
    asm MOV ES,BX
    asm XOR BX,BX
    asm ADD BX,portnum
    asm MOV AX, WORD PTR ES:[BX]
    asm MOV baseport,AX

    if(portnum == SER_P1)
    {
		SER_baseport1 = baseport;
        HardwareFlowControl1 = HWDFlowControl;
    }
    else if(portnum == SER_P2)
    {
       HardwareFlowControl2 = HWDFlowControl;
	   SER_baseport2 = baseport;
    }

    SetBaudRate(baseport, brate);
    SetWordLength(baseport);

    ClearLSR_MSR_IIR(baseport);

    /* Insane measure. Make sure that 7th bit of LCR is 0 */
    UnSetControlBit(baseport);
    /* Read and clear any left over characters */
    asm MOV DX,baseport
LOOP:
	asm ADD DX,5
    /* Get line status */
	asm IN AL,DX
    /* Check if char available */
	asm TEST AL,01H
    asm JZ CHARNOTFOUND
    asm SUB DX,5
	asm IN AL,DX
    asm JMP LOOP  /* Are all available characters read? */
CHARNOTFOUND:
    ClearAndEnableFIFO(baseport, actions->Chars_Bef_Irq);
	/* Deal with interrupts */
	if(actions)
	{
		/* Set interrupt handler */
		if(portnum == SER_P1)
		{
			IrqHandlerForCom1_Changed = 1;
			SER1_func = actions->IrqActions;
            IrqHandler_For_Com1 = (void interrupt (*)())(getvec(IRQVEC_FOR_SER_P1));
			setvec(IRQVEC_FOR_SER_P1, InterruptHandlerForCOM1);
			/* Disable flag 4 in 8259 */
			asm MOV DX,21H
			asm IN AL,DX
			asm AND AL,0EFH
			asm OUT DX,AL
		}
		else
		{
			IrqHandlerForCom2_Changed = 1;
            SER2_func = actions->IrqActions;
            IrqHandler_For_Com2 = (void interrupt (*)())(getvec(IRQVEC_FOR_SER_P2));
            setvec(IRQVEC_FOR_SER_P2, InterruptHandlerForCOM2);
			/* Disable flag 3 in 8259 */
			asm MOV DX,21H
			asm IN AL,DX
			asm AND AL,0F7H
			asm OUT DX,AL
		}

		/* NrOfCharsBefReadIRQ(baseport, actions->Chars_Bef_Irq);*/
		IER_Enable(baseport, actions->IR_Byte);
		MCR_Enable(baseport);
	}
	else
	{
		if(portnum == SER_P1)
			IrqHandlerForCom1_Changed = 0;
		else if(portnum == SER_P2)
			IrqHandlerForCom2_Changed = 0;

        IER_Disable(baseport);
	}

    if(HWDFlowControl)
    {
        /* Setting RTS to 0*/
        asm MOV DX,baseport
        asm ADD DX,4
        asm IN AL,DX
        asm AND AL,0FDH
        asm OUT DX,AL
        /* Make 0 the counters */
        CharsSent1 = MAXCHARS_BEFORE_STOPPING;
        CharsSent2 = MAXCHARS_BEFORE_STOPPING;
        PrevSignal1 = 0;
        PrevSignal2 = 0;
     }
     asm POPF
     /*     THR_Empty interrupts,      */
	 /*    ,  UART      interrupt */
	 /*       (   )  */
	 /*             . */
	 /*        interrupt  ,   */
	 /*      UART,       */
	 /*  interrupts.          IIR,*/
	 /*  acknowledge sto UART. */

     InterruptReason(portnum);
}

/* H        polling     */
/* interrupt handlers.         */
/*   UART      ReadCh,        */
/* 16 .       .      */
/*  0.      -2,       */
/*  buffers  UART       (overwrite error)  -1 */

KERNEL_INT SER_ReadChar(SER_PORT portnum, char *ReadCh)
{
        int baseport;
        char ans;
        int mret = 0;
        char *ptr = ReadCh;
        unsigned char HardwareFlowControl;

        if(portnum == SER_P1)
        {
                HardwareFlowControl = HardwareFlowControl1;
				baseport = SER_baseport1;
        }
        else if(portnum == SER_P2)
        {
                HardwareFlowControl = HardwareFlowControl2;
				baseport = SER_baseport2;
        }
        else
                return -10; /* Not valid serial port */

        if(HardwareFlowControl)
        {
                /* Characters read. Send signal to disable sending of more chars*/
                asm MOV DX,baseport
                asm ADD DX,4
                asm IN AL,DX
                asm AND AL,0FDH /* make RTS 0 */
                asm OUT DX,AL
        }
LOOP:
        /* Check the corresponding bit of the control status register */
		asm MOV DX,baseport
		asm ADD DX,5
        /* Get line status */
		asm IN AL,DX
        asm TEST AL,08H
		asm JNZ TR_ERROR
		asm TEST AL,02H
		asm JNZ CHAR_OVERFLOW
		asm TEST AL,01H
        asm JZ CHARNOTFOUND
        asm SUB DX,2
		asm IN AL,DX
		asm AND AL,7FH
		asm OUT DX,AL
		asm SUB DX,3
		asm IN AL,DX
        asm MOV ans,AL
        *ptr++ = ans;
        mret++;
        asm JMP LOOP  /* Are all available characters read? */

        /* Hardware Flow control */
        if(HardwareFlowControl && mret == 0) /* If we have hardware flow control and no characters read */
        {
                /* Checking for signal */
                asm MOV DX,baseport
                asm ADD DX,6
                asm IN AL,DX
                asm TEST AL,10H
                asm JZ CHARNOTFOUND /* An den eixame shma gia flow control apla synexizoume */
                /* Sending signal for permission */
                asm MOV DX,baseport
                asm ADD DX,4
                /* send signal for more data */
                asm IN AL,DX
                asm OR AL,02H
                asm OUT DX,AL
        }

CHARNOTFOUND:
        return mret; /* If mret == 0, No char returned */
CHAR_OVERFLOW:
        return -1; /* Character overwritten at the serial buffers */
TR_ERROR:
        return -2; /* Transmission error */
}

/*          UART.     */
/*     polling     interrupt handlers.     */
/* 1      .  -2   transmission error.   */
/* -1    buffers  UART           */
/* .   2       .    */
/*            <> 1 */

KERNEL_INT SER_WriteChar(SER_PORT portnum, char WriteCh)
{
        int baseport, retval = 1;
        unsigned char HardwareFlowControl;
        unsigned char *PrevSignal, CurSignal;
        unsigned int *CharsSent;

        if(portnum == SER_P1)
        {
                HardwareFlowControl = HardwareFlowControl1;
                PrevSignal = &PrevSignal1;
                CharsSent = &CharsSent1;
				baseport = SER_baseport1;
        }
        else if(portnum == SER_P2)
        {
                HardwareFlowControl = HardwareFlowControl2;
                PrevSignal = &PrevSignal2;
                CharsSent = &CharsSent2;
				baseport = SER_baseport2;
        }
        else return -10; /* Not valid serial port */

        /* Hardware flow control check (if necessary) */
        if(HardwareFlowControl)
        {
                /* Checking for permission */
                asm MOV DX,baseport
                asm ADD DX,6
                asm IN AL,DX
                asm AND AL,10H
                asm MOV CurSignal,AL
                if(*PrevSignal !=  CurSignal) /* Change in state. Clear to send */
                {
                        *CharsSent = 1;
                        goto SENDCHAR;
                }
                else if(CurSignal == *PrevSignal && *CharsSent < MAXCHARS_BEFORE_STOPPING)
                {
                        (*CharsSent)++;
                        goto SENDCHAR;
                }
                else
                {
                        /* Send signal for permission */
                        asm MOV DX,baseport
                        asm ADD DX,4
                        /* intention to send data */
                        asm IN AL,DX
                        asm OR AL,02H
                        asm OUT DX,AL
                        return 3; /* sent signal and waiting for permission */
                }
        }
SENDCHAR:
		asm MOV DX,baseport
		asm ADD DX,5
		/* Get line status */
		asm IN AL,DX
		asm TEST AL,08H
		asm JNZ TR_ERROR
        /* Check THR (Transmitter holding register) */
		asm TEST AL,20H
	    asm JZ THR_NOT_EMPTY
		asm TEST AL,40H
        asm JZ TSR_NOT_EMPTY
CONT:	asm SUB DX,2
		asm IN AL,DX
		asm AND AL,7FH /* Set 7th bit of line control to 0 */
		asm OUT DX,AL
		asm SUB DX,3
		asm MOV AL,BYTE PTR WriteCh
		asm OUT DX,AL
        return retval;
TSR_NOT_EMPTY:
        return 2;
        /* data are being sent at this very moment and have not reached the recipient yet from */
        /* the previous char */
THR_NOT_EMPTY:
        return -1; /* There are data waiting to be sent in THR */
TR_ERROR:
        return -2; /* Transmission error */
}

/* H       interrupt handler       */
/*        .   IIR */
/* Returns the cause of the interrupt (used only inside interrupt handlers) */
/* Returns 0 if the specified port hasn't triggered the interrupt */
/* Returns 1 if new char is available */
/* Returns 2 if a new char can be sent (THR Empty) */
/* Returns 3 if modem status has changed */
/* Returns -1 if a transmission error occured */

KERNEL_INT InterruptReason(SER_PORT port)
{
        int baseport;
        if(port == SER_P1)
                baseport = SER_baseport1;
        else
                baseport = SER_baseport2;
        /* Checking the serial IIR to determine what caused the interrupt */
        asm MOV DX,baseport
		asm ADD DX,2
		asm IN AL,DX
		asm TEST AL,01H
		/* If bit 0 is 1 then another port triggered the interrupt */
        asm JNZ end
cause:
		asm TEST AL,02H
		asm JNZ THR_NOT_EMPTY
		asm TEST AL,04H
		asm JNZ CHAR_READY
		asm TEST AL,06H
		asm JNZ TRANSM_ERROR
		/* diaforetika einai status change */
		return 3;

THR_NOT_EMPTY:
		return 2;
CHAR_READY:
		return 1;
TRANSM_ERROR:
		return -1;
end:
		return 0;
}

/* ---------------------------------------------------------------- */
/* A    (interrut driven)           */
/* ---------------------------------------------------------------- */

/*     buffers */
static int Distance(char *grt, char *lst)
{
        return (int)(grt >= lst ? grt - lst : BUFLEN - (lst - grt));
}

/*       interrupt handlers      */

static KERNEL_VOID Handler(SER_PORT port, char *InBuf, char **SER_InBufPtrWr, char **SER_InBufPtrRd, int ISEM, int OSEM)
{
	char buf[16];
	int i, ans;
	/* Find out what happened */
    ans = InterruptReason(port);
    switch(ans)
	{
		case 1: /* new char available */
			ans = SER_ReadChar(port, buf);
			if(ans < 0)
			{
				QSR_Error.Error_Type = 1;
				QSR_Error.Error = ans;
                                return;
			}
			else
				QSR_Error.Error_Type = 0;

			for(i = 0; i < ans; i++)
			{
                 if(Distance(*SER_InBufPtrWr, *SER_InBufPtrRd) >= BUFLEN - 1)
                 {
					 /* If we have exceeded our buffer capacity, discard the next chars */
                     QSR_Error.Error_Type = 3;
                     return;
                  }
				  **SER_InBufPtrWr = buf[i];
                  (*SER_InBufPtrWr)++;
                  if(*SER_InBufPtrWr - InBuf >= BUFLEN)
				  {
					*SER_InBufPtrWr = InBuf;
					*InBuf = buf[i];
				  }
                  V(ISEM);
            }
			break;
        case 2:
                  V(OSEM); /* Give Permission to write */
			break;
		default:
			break;
	}
}

static KERNEL_VOID Handler1(void)
{
        Handler(SER_P1, SER1_InBuf, &SER1_InBufPtrWr, &SER1_InBufPtrRd, ISEM1, OSEM1);
}

static KERNEL_VOID Handler2(void)
{
        Handler(SER_P2, SER2_InBuf, &SER2_InBufPtrWr, &SER2_InBufPtrRd, ISEM2, OSEM2);
}

KERNEL_VOID QSR_InitializePort(SER_PORT portnum, BAUD_RATE brate)
{
	struct IrqConf mconf;
	int i;

	if(portnum == SER_P1)
		mconf.IrqActions = Handler1;
        else
		mconf.IrqActions = Handler2;

	/* trigger interrupts when a new char is ready or when a character can be sent */
        mconf.IR_Byte = IRQ_When_Char_Ready | IRQ_When_THR_Empty;
	/* trigger interrupt with every character coming */
	mconf.Chars_Bef_Irq = ch1;

    asm PUSHF
    asm CLI
    SER_InitializePort(portnum, brate, 0, &mconf);

	if(portnum == SER_P1)
	{
        SER1_InBufPtrRd = SER1_InBuf;
        SER1_InBufPtrWr = SER1_InBuf;
		V(OSEM1);
	}
	else
	{
        SER2_InBufPtrRd = SER2_InBuf;
        SER2_InBufPtrWr = SER2_InBuf;
		V(OSEM2);
	}
        asm POPF
}


KERNEL_CHAR QSR_ReadChar(SER_PORT portnum)
{
	char ch = 0;
	if(portnum == SER_P1)
	{
		/*           */
		/*   buffer  SER1_InBufPtrRd */
        P(ISEM1);
		ch = *SER1_InBufPtrRd++;
		if(SER1_InBufPtrRd - SER1_InBuf >= BUFLEN)
			SER1_InBufPtrRd = SER1_InBuf;
	}
	else
	{
		/*           */
		/*   buffer  SER2_InBufPtrRd */
        P(ISEM2);
		ch = *SER2_InBufPtrRd++;
		if(SER2_InBufPtrRd - SER2_InBuf >= BUFLEN)
			SER2_InBufPtrRd = SER2_InBuf;
	}
	return ch;
}

KERNEL_INT QSR_WriteChar(SER_PORT portnum, char ch)
{
	int ans;
    /* Block for permission from the interrupt handler */
	/*      UART     */
	/*      .    */
	/*    interrupt handler */
	if(portnum == SER_P1)
		P(OSEM1);
	else if(portnum == SER_P2)
		P(OSEM2);
	else /* should not happen */
		return 10;

	ans = SER_WriteChar(portnum, ch);
	return ans;
}

KERNEL_VOID QSR_FinalizePorts(void)
{
        RestoreInterrupts();
}

/*==========================================================================================*/
/*=====================  O (MEMORY MAPPED) =====================================*/
/*==========================================================================================*/

/* KERNEL_VOID PrintChar (int x, int y, char content):    */
/* int x, int y:     x,y */
/* char content:     */

KERNEL_VOID PrintChar (int x, int y, char content)
{
	int pos;
	pos=y*160+2*x;
	asm mov bx,pos;
	asm mov ax,0B800h ;
	asm mov es,ax;
	asm mov ah,00Fh ;
	asm mov al, content;
	asm mov es: [bx],ax ;
}

/*   string   x,y.   length  */

KERNEL_VOID PrintString (int x, int y, char *str,int length)
{
	int i;
	for (i=0;i<length;i++)
		PrintChar (x+i,y,str[i]);
}

/*   string   x,y.     */
/*    NULL */

KERNEL_VOID PrintStr(int x, int y, char *string)
{
	char *ptr;

	for(ptr = string; *ptr; ptr++)
		PrintChar(x+(ptr - string), y, *ptr);
}

/*      x,y */

KERNEL_VOID PrintInteger (int x, int y, int number)
{
	int remainder;
	int divisor;
	int pos;
	char temp;

	remainder=number;
	divisor=10000;
	pos=x;
	while (divisor>0)
	{
		temp=remainder/divisor;
		remainder=remainder-temp*divisor;
		divisor=divisor/10;
		/* Now temp has the appropriate decimal digit */
		PrintChar (pos,y,0x30+temp);
		pos ++;
	}
}

/*      x,y */

KERNEL_VOID PrintHex(int x, int y, unsigned int number)
{
    int pos, shift;
	char temp;
	unsigned int mask;

    mask = 0xF000;
    pos = x;
    shift = 12;
    while(mask)
    {
		temp = (number & mask) >> shift;
        mask >>= 4;
        shift -= 4;
        if(temp > 9)
			PrintChar(pos, y, 'a' + temp - 10);
        else
            PrintChar(pos, y, '0' + temp);
        pos++;
    }
}

/*    */

KERNEL_VOID cls(void)
{
	int i;
	int j;
	for (i=0;i<80;i++)
		for (j=0;j<24;j++)
			PrintChar (i,j,' ');
}

/*==========================================================================================*/
/*==============================  O (BIOS) =====================================*/
/*==========================================================================================*/

/*                 */
/*       BIOS  .  High Intensity   */
/*   flag <> 0 */

KERNEL_VOID bSetColor(SCR_COLORS fgcolor, SCR_COLORS bgcolor, unsigned int HighIntensity)
{
        /* Disable interrupts because someone may try to read it inside an interrupt handler
        in an unsafe state (although not recommended) */
        asm PUSHF
        asm CLI
        SCR_Current_Color = ((bgcolor << 4) | fgcolor) & 0xFF;
        if(HighIntensity)
        	SCR_Current_Color |= 0x8;
        asm POPF
}

/*   cursor        10
/*   2 */

KERNEL_VOID bSetCursor(int x, int y)
{
        asm MOV AH,2
        asm MOV DH,y
        asm MOV DL,x
        asm MOV BH,0
#if SIRQS_ISOLATED == 1
		asm PUSHF
		asm CLI
#endif
        asm INT 10H
#if SIRQS_ISOLATED == 1
		asm POPF
#endif
}

/*       cursor  ,    */
/*  10   3 */

KERNEL_VOID bGetCursor(int *x, int *y)
{
        if(x && y)
        {
			asm MOV AH,3
            asm MOV BH,0
#if SIRQS_ISOLATED == 1
			asm PUSHF
			asm CLI
#endif
            asm INT 10H
#if SIRQS_ISOLATED == 1
			asm POPF
#endif

            *x = _DL;
            *y = _DH;
        }
}

/*       10   6 */

KERNEL_VOID bcls(void)
{
        asm MOV AH,6
        asm MOV AL,0
        asm MOV CH,0
        asm MOV CL,0
        asm MOV DH,24
        asm MOV DL,79
        asm MOV BH,07H /*   default  */
#if SIRQS_ISOLATED == 1
		asm PUSHF
		asm CLI
#endif
        asm INT 10H
#if SIRQS_ISOLATED == 1
		asm POPF
#endif
        /*      0,0 */
        asm MOV AH,2
        asm XOR DX,DX
        asm XOR BH,BH
#if SIRQS_ISOLATED == 1
		asm PUSHF
		asm CLI
#endif
        asm INT 10H
#if SIRQS_ISOLATED == 1
		asm POPF
#endif
}

/*   cursor     .   cursor      (column 79) */
/*         .   cursor     ,  */
/*   scroll                */
/*     */

KERNEL_VOID bAdvanceCursor(void)
{
		/*  : 25 X 80 */
        asm MOV AH,3
        asm MOV BH,0
#if SIRQS_ISOLATED == 1
		asm PUSHF
		asm CLI
#endif
        asm INT 10H
#if SIRQS_ISOLATED == 1
		asm POPF
#endif

        asm MOV AH,2
        asm INC DL
        /*    */
        asm CMP DL,80
        asm JB cont1
        asm XOR DL,DL
        asm INC DH
        /*    */
        asm CMP DH,25
        asm JB cont1
        /* scroll up    */
        asm PUSH AX
        asm PUSH DX
        asm MOV AH,6
        asm MOV AL,1
        asm MOV CH,0
        asm MOV CL,0
        asm MOV DH,24
        asm MOV DL,79
        asm MOV BH,SCR_Current_Color
#if SIRQS_ISOLATED == 1
		asm PUSHF
		asm CLI
#endif
        asm INT 10H
#if SIRQS_ISOLATED == 1
		asm POPF
#endif
        asm POP DX
        asm POP AX
        asm DEC DH
        asm XOR BH,BH
cont1:
#if SIRQS_ISOLATED == 1
		asm PUSHF
		asm CLI
#endif
		asm INT 10H
#if SIRQS_ISOLATED == 1
		asm POPF
#endif
}

/*    ,     cursor */

KERNEL_VOID bBacktrackCursor(void)
{

		/*  : 25 X 80 */
        asm MOV AH,3
        asm MOV BH,0

#if SIRQS_ISOLATED == 1
		asm PUSHF
		asm CLI
#endif
        asm INT 10H
#if SIRQS_ISOLATED == 1
		asm POPF
#endif

        asm MOV AH,2
        /*    */
        asm CMP DL,0
        asm JG cont1
        asm MOV DL,80
        asm CMP DH,0
        asm JG cont2
        /* scroll down    */
        asm PUSH AX
        asm PUSH DX
        asm MOV AH,6
        asm MOV AL,2
        asm MOV CH,0
        asm MOV CL,0
        asm MOV DH,24
        asm MOV DL,79
        asm MOV BH,SCR_Current_Color

#if SIRQS_ISOLATED == 1
		asm PUSHF
		asm CLI
#endif
        asm INT 10H
#if SIRQS_ISOLATED == 1
		asm POPF
#endif

        asm POP DX
        asm POP AX
        asm XOR BH,BH
cont2:  asm DEC DH
cont1:  asm DEC DL

#if SIRQS_ISOLATED == 1
		asm PUSHF
		asm CLI
#endif
		asm INT 10H
#if SIRQS_ISOLATED == 1
		asm POPF
#endif
}

/*  .          scroll */
/*           cursor      */
/*   */

KERNEL_VOID bNewLine(void)
{
        asm MOV AH,3
        asm MOV BH,0
#if SIRQS_ISOLATED == 1
		asm PUSHF
		asm CLI
#endif
        asm INT 10H
#if SIRQS_ISOLATED == 1
		asm POPF
#endif

        asm MOV AH,2
        asm INC DH
        /*    */
        asm CMP DH,25
        asm JB cont1
        /* scroll up    */
        asm PUSH AX
        asm PUSH DX
        asm MOV AH,6
        asm MOV AL,1
        asm MOV CH,0
        asm MOV CL,0
        asm MOV DH,24
        asm MOV DL,79
        asm MOV BH,SCR_Current_Color
#if SIRQS_ISOLATED == 1
		asm PUSHF
		asm CLI
#endif
        asm INT 10H
#if SIRQS_ISOLATED == 1
		asm POPF
#endif
        asm POP DX
        asm POP AX
        asm DEC DH
cont1:  asm XOR DL,DL
        asm XOR BH,BH
#if SIRQS_ISOLATED == 1
		asm PUSHF
		asm CLI
#endif
        asm INT 10H
#if SIRQS_ISOLATED == 1
		asm POPF
#endif
}

/*    c       cursor      */
/*        (    AdvanceCursor).   */
/*   (   attribute)       bSetColor */

KERNEL_VOID bPrintChar(char c)
{
        /*   */
        asm MOV AH,9h
        asm MOV AL,c
        asm MOV BH,0H
        asm MOV BL,SCR_Current_Color
        asm MOV CX,1
#if SIRQS_ISOLATED == 1
		asm PUSHF
		asm CLI
#endif
        asm INT 10H
#if SIRQS_ISOLATED == 1
		asm POPF
#endif

        asm MOV AH,3
        asm MOV BH,0
#if SIRQS_ISOLATED == 1
		asm PUSHF
		asm CLI
#endif
        asm INT 10H
#if SIRQS_ISOLATED == 1
		asm POPF
#endif

        asm MOV AH,2
        asm INC DL
        /*    */
        asm CMP DL,80
        asm JB cont1
        asm XOR DL,DL
        asm INC DH
        /*    */
        asm CMP DH,25
        asm JB cont1
        /* scroll up    */
        asm PUSH AX
        asm PUSH DX
        asm MOV AH,6
        asm MOV AL,1
        asm MOV CH,0
        asm MOV CL,0
        asm MOV DH,24
        asm MOV DL,79
        asm MOV BH,SCR_Current_Color
#if SIRQS_ISOLATED == 1
		asm PUSHF
		asm CLI
#endif
        asm INT 10H
#if SIRQS_ISOLATED == 1
		asm POPF
#endif
        asm POP DX
        asm POP AX
        asm DEC DH
        asm XOR BH,BH
cont1:
#if SIRQS_ISOLATED == 1
		asm PUSHF
		asm CLI
#endif
		asm INT 10H
#if SIRQS_ISOLATED == 1
		asm POPF
#endif
}

/*  string       .   length */
/*  */

KERNEL_VOID bPrintString (char *str,int length)
{
	int i;
	for (i=0;i<length;i++)
		bPrintChar(str[i]);
}

/*  string       .   */
/*     NULL  */

KERNEL_VOID bPrintStr(char *string)
{
	char *ptr;

	for(ptr = string; *ptr; ptr++)
		bPrintChar(*ptr);
}

/*    x */

KERNEL_VOID bPrintInteger (int x)
{
	int remainder;
	int divisor;
	char temp;
	int start = 1;

	if(x == 0)
	{
		bPrintChar('0');
		return;
	}

	if(x < 0)
		bPrintChar('-');

	remainder=x<0 ? x*(-1) : x;
	divisor=10000;
	while (divisor>0)
	{
		temp=remainder/divisor;
		remainder=remainder-temp*divisor;
		divisor=divisor/10;  /* T  temp      */
		if(start != 1 || temp != 0)
		{
			bPrintChar('0'+temp);
			start = 0;
		}
	}
}

/* UpDown = 0, means down, otherwise means up */
/* Scrolling         (UpperLeftX, UpperLeftY) & (LowerRightX, LowerRightY). */
/*    -       (0,0) */
/* UpDown = 0, means down, otherwise means up */

KERNEL_VOID bScrollWindow(const KERNEL_CHAR UpperLeftX, const KERNEL_CHAR UpperLeftY, const KERNEL_CHAR LowerRightX, const KERNEL_CHAR LowerRightY, const KERNEL_CHAR UpDown)
{
        /* scroll window up */
        if(UpDown)
			asm MOV AH,6
        else
            asm MOV AH,7
        asm MOV AL,1
        asm MOV CH,BYTE PTR UpperLeftX
        asm MOV CL,BYTE PTR UpperLeftY
        asm MOV DH,BYTE PTR LowerRightX
        asm MOV DL,BYTE PTR LowerRightY
        asm MOV BH,07H

#if SIRQS_ISOLATED == 1
		asm PUSHF
		asm CLI
#endif
        asm INT 10H
#if SIRQS_ISOLATED == 1
		asm POPF
#endif
}

/*==========================================================================================*/
/*==============================   (BIOS) ===============================*/
/*==========================================================================================*/

/* H            BIOS 16H ( 0).  */
/*   ExtendedCode   NULL,    extended code,    */
/*  extended code     */

char bGetChar(unsigned char *ExtendedCode)
{
	char mret;
	unsigned char xx;
	asm XOR AH,AH

#if SIRQS_ISOLATED == 1
	asm PUSHF
	asm CLI
#endif
	asm int 16H
#if SIRQS_ISOLATED == 1
	asm POPF
#endif

	asm MOV mret,AL
	asm MOV xx, AH

	if(ExtendedCode) *ExtendedCode = xx;

	return mret;
}


/*==========================================================================================*/
/*==============================   ==========================================*/
/*==========================================================================================*/

/*          int *ps */

KERNEL_VOID disabl_intr(int *ps)
{
int tmp;
asm pushf ;
asm mov al,0FFh ;
asm out 21h,al ;
asm out 0A1h, al ;
asm pop tmp
*ps = tmp;
}

/*        int *ps	*/

KERNEL_VOID restor_intr(int *ps)
{
int tmp;
tmp = *ps;
asm push tmp ;
asm popf
}

/*    	*/

KERNEL_VOID enable_all_intr()
{
asm cli
asm xor al,al
asm out 21h,al
asm out 0A1h,al
asm sti
}

/*                  */
/* .        IER    mask  IRQs */
/* 3,4  8259.      interrupts handlers   timer,  IRQs 3,4  */
/*  interrupt 28H (DOS IDLE LOOP) */

KERNEL_VOID RestoreInterrupts(void)
{
	asm CLI

	/* Disable interrupt generation at MCR */
	if(IrqHandlerForCom1_Changed)
	{
		asm MOV DX,SER_baseport1
		asm ADD DX,4
		asm IN AL,DX
		asm AND AL,0F7H
		asm OUT DX,AL
        /* Clear pending interrupts */
        ClearLSR_MSR_IIR(SER_baseport1);
	}

	if(IrqHandlerForCom2_Changed)
	{
		asm MOV DX,SER_baseport2
		asm ADD DX,4
		asm IN AL,DX
		asm AND AL,0F7H
		asm OUT DX,AL
        /* Clear pending interrupts */
         ClearLSR_MSR_IIR(SER_baseport2);
	}

	/* Disable interrupts 3 & 4 at 8259 interrupt controller. We should normally change to the previous status */
	asm MOV DX,21H
	asm IN AL,DX
	asm OR AL,18H
	asm OUT DX,AL
	/* Restore the previous interrupts vectors */
	if(IrqHandlerForCom1_Changed)
		setvec(IRQVEC_FOR_SER_P1, IrqHandler_For_Com1);
	if(IrqHandlerForCom2_Changed)
                setvec(IRQVEC_FOR_SER_P2, IrqHandler_For_Com2);

    /* Restore interrupt 28H */
    asm XOR AX,AX
    asm MOV ES,AX
    asm MOV BX,ds_store
    asm MOV AX, WORD PTR OldInt28
    asm MOV WORD PTR ES:[BX],AX

    SER_baseport1 = SER_baseport2 = 0; /* These are the indicators that we have initialized the */
    								   /* corresponding port */
    /* We must set them to 0, so that the SER_InitializePort function initializes them again */

    /* Restore Clk interrupt handler */
    if(IrqHandler_For_Clk)
         setvec(0x08, IrqHandler_For_Clk);

	asm STI
}

/*==========================================================================================*/
/*==============================   ==========================================*/
/*==========================================================================================*/

/* H  setclkrate       */

static KERNEL_VOID setclkrate()
{
	unsigned long count;
	unsigned low_byte, high_byte;
	int ps;

	disabl_intr(&ps); /*    */

	count = (unsigned long)(23863);
	low_byte  = (unsigned)(count & 0xFF);
	high_byte = (unsigned)((count >> 8) & 0xFF);
	port_out(0x43, 0, 0x36);
	port_out(0x40, 0, low_byte);
	port_out(0x40, 0, high_byte);

	restor_intr (&ps); /*     */
	enable_all_intr(); /*    */
}

/*   clkinit       */

static KERNEL_VOID interrupt clkint()
{
	asm push ax;	/*       */
	asm push dx;
	asm push ds;
	asm push bx;
	asm push cx;
	asm push si;
	asm push di;
	asm push es;

	asm mov  al, 20h; /*    8159    */
    asm mov  dx, 20h; /*  */
    asm out  dx,al;

	preempt--;		/*     preemt   */
    if (preempt==0)    /*  ,       */
   		reschedule (); /*  reschedule */
	asm pop es;		/*    */
	asm pop di;
	asm pop si;
	asm pop cx;
	asm pop bx;
	asm pop ds;
	asm pop dx;
	asm pop ax;
}

/*   clkinit     ..,   */
/*   . */

static KERNEL_VOID clkinit()
{
	IrqHandler_For_Clk = (void interrupt (*)())getvec(0x08); /*       */
	setvec(0x08,clkint); /*        */
	setclkrate(); /*      */
	preempt = QUANTUM; /*     preempt */
}



/*==========================================================================================*/
/*==============================   ========================================*/
/*==========================================================================================*/

/*          .. */
/* void *procaddr:          */
/* int priority:     */
/* char *name:     */
/* int type:     */

KERNEL_VOID create (void *procaddr, int priority, char *name, int type)
{
	int ps;
	int i;

	disabl_intr(&ps); /*    */
	numproc++;		  /*        */
	proctab[numproc].pid=numproc; 					/*	 PID   */
	proctab[numproc].pstate=READY;					/*    */
	for (i=0;(i<PNMLEN-1) && (name[i]!='\0');i++)   /*     */
            proctab[numproc].pname[i]=name[i];
	proctab[numproc].pname[i]='\0';
	proctab[numproc].pprio=priority;				/*     */
	proctab[numproc].type=type;						/*     */

	for (i=0;i<14;i++)								/*    0 */
		proctab[numproc].pregs[i]=0;

    proctab[numproc].pregs[RIP]  = getoffset(enable_all_intr);  /* The offset of the function address  -> IP */
    proctab[numproc].pregs[RCS]  = getsegment(enable_all_intr); /* The segment of the function address -> CS */
	proctab[numproc].pregs[PSW] = INITPS;			   /*     */
	/* The offset of the stack (normally it would be -1 but we insert data inside */
	/* the stack in the following line) */
    proctab[numproc].pregs[RSP]  = getoffset (&stacks[numproc][STACKSIZE-2]);
    stacks[numproc][STACKSIZE-2]= getoffset(procaddr);
    stacks[numproc][STACKSIZE-1]= getsegment(procaddr);
    proctab[numproc].pregs[RSS]  = getsegment (stacks); /* The segment of the stack */
    proctab[numproc].pregs[RDS]  = _DS; /* The data segment is common for all processes */

	restor_intr (&ps); /*     */
	enable_all_intr(); /*    */

}

/*   reschedule    .     */
/*               */

KERNEL_INT reschedule ()
{
	int previous;
	int ps;

	disabl_intr(&ps); /*    */
	preempt=QUANTUM;

	/*             */
	/* --------------------------------------------------------------------------- */

	if (current==-1) { /*         reschedule() */
	 previous=0; /*     PID=0 */
	 current=1;  /*       PID=1 */
 	}
	else { /*  */
	 previous=current; /*      */
	 current++; /*       */
	 if (current>numproc) current=1; /*          */
	 while (proctab[current].pstate==BLOCKED) { /*     blocked */
	  current++; /*    */
	  if (current>numproc) current=1; /*          */
	 }
	}
	/*----------------------------------------------------------------------------- */

	if (proctab[previous].pstate==RUNNING) proctab[previous].pstate=READY;
	proctab[current].pstate=RUNNING;

	/*   */
	ctxsw (proctab[previous].pregs,proctab[current].pregs);

	restor_intr (&ps); /*     */
	enable_all_intr(); /*    */

	return OK;
}

/*==========================================================================================*/
/*==============================   ======================================*/
/*==========================================================================================*/

/*         P     */
/*   int s */

KERNEL_VOID P(int s)
{
	int ps;
	disabl_intr(&ps); /* A   */

	if (semavalue[s]>0)   /*       >0 */
		semavalue[s]--;   /*   */
	else	{			  /*  */
		proctab[current].pstate=BLOCKED; /*    */
		semaphores[s][current]=1;        /*       s */
        reschedule();					 /*    */
	}

	restor_intr (&ps); /*     */
	enable_all_intr(); /*    */
}

/*         S     */
/*   int s */
KERNEL_VOID V(int s)
{
	int ps;
	int i;
	disabl_intr(&ps); /* A   */
	i=0;
	/*         s */
	while ((semaphores[s][i]!=1)&&(i<=numproc))
		i++;
	if (i==numproc+1) semavalue[s]++; /*    ,     */
	else {
		semaphores[s][i]=0; /*        */
		proctab[i].pstate=READY; /*       */
	}

	restor_intr (&ps); /*     */
	enable_all_intr(); /*    */
}



/*=========================================================================================*/
/*    null process.        */
/*   .. */

void Idle(void) {
 while(1) ;
}

/*=========================================================================================*/
/*        ..   */
/*   */

void Initialize (void) {
	int i,j;
	for (i=0;i<MAXSEM;i++)				/* A   */
		for (j=0;j<MAXPROC;j++) {
			semaphores[i][j]=0;
			semavalue[i]=0;
		}
	bSetColor(SCR_White, SCR_Black, 0); /*     */
	numproc=0;							/*     . = 0 */
	current=-1;							/*      */
}

/*=========================================================================================*/
/*        ..     */
/*      .. */
static KERNEL_VOID interrupt dummy()
{
	asm push ax;	/*       */
	asm push dx;
	asm push ds;
	asm push bx;
	asm push cx;
	asm push si;
	asm push di;
	asm push es;

	asm mov  al, 20h; /*    8159    */
    asm mov  dx, 20h; /*  */
    asm out  dx,al;

	bSetCursor (40,11);
	bPrintStr("DOS !!!!!!");

	asm pop es;		/*    */
	asm pop di;
	asm pop si;
	asm pop cx;
	asm pop bx;
	asm pop ds;
	asm pop dx;
	asm pop ax;
}

void main(void)
{
	int ps;
	disabl_intr(&ps);
	setvec(0x21,dummy);
	Initialize(); /*    .. */
    create(Idle,0,"Idle",USER_PROCESS); /*  null process */

	/*======================================================================*/
	/*==================    ======================*/
	/*======================================================================*/
	Initializer();

	create (P1,0,"P1",USER_PROCESS);
	create (P2,0,"P2",USER_PROCESS);

	/*      */


	/*======================================================================*/
	/*======================================================================*/

	clkinit(); /*       */
	restor_intr (&ps);
	enable_all_intr();
	while (1);
}
/*=========================================================================================*/
