/*-----------------------------------------------------*/
/*--	Micro-EMPIX CORE  (Micro.h)                   -*/
/*--    Ver 1.0 (c) 2004                              -*/
/*--	E             -*/
/*--    ==================================            -*/
/*--    .                  -*/
/*--    .                        -*/
/*--    . .                      -*/
/*-----------------------------------------------------*/

/* ====================================================*/
/* ==========GLOBAL DEFINITIONS FOR OS FEATURES =======*/
/* ====================================================*/

#define QUANTUM 2               /* The time quantum given to each process */
#define MAXPROC 5 			/* The maximum number of processes allowed */
#define MAXSEM  20
#define STACKSIZE 2000 	/* The size of the stack for each process */

#define NULL 0

/* ====================================================*/
/* ==========INTERNAL DEFINITIONS FOR PROCESSES =======*/
/* ====================================================*/

#define RBP         0
#define RAX         1
#define RBX         2
#define RCX         3
#define RDX         4
#define RSI         5
#define RDI         6
#define RDS         7
#define RSS         8
#define RES         9
#define RSP         10
#define RIP         11
#define RCS         12
#define PSW        13
#define INITPS     0x0200
#define KERNEL_CHAR char
#define USER_INT int
#define KERNEL_PROCESS 0
#define USER_PROCESS 1
#define RESCHEDULE_DUE_TO_CLOCK 0
#define RESCHEDULE_DUE_TO_FCALL 1
#define OK 1
#define ERROR 0
#define RUNNING 'R'
#define BLOCKED 'B'
#define READY 'A'
#define PNMLEN 15
#define PNREGS 14
#define KERNEL_INT int
#define KERNEL_VOID void
#define KERNEL_CHAR char

/* ====================================================*/
/* ==========   DEFINITIONS RELATED TO I/O ============*/
/* ====================================================*/

typedef enum SCR_COLORS
{
        SCR_White = 0x7,
        SCR_Black = 0x0,
        SCR_Red = 0x4,
        SCR_Green = 0x2,
        SCR_Yellow = 0x6,
        SCR_Cyan = 0x3,
        SCR_Purple = 0x5,
        SCR_Blue = 0x1
} SCR_COLORS;

/* For chars_bef_irq */
#define ch1 0x00
#define ch4 0x40
#define ch8 0x80
#define ch14 0xC0

/* for IR_Byte: For example: Enable interrupts for reading and writing for every incoming/outgoing char: */
/* struct IrqConf *x = malloc(sizeof(struct IrqConf)); */
/* x->IR_Byte = IRQ_When_Char_Ready | IRQ_When_THR_Empty; */
/* x->Chars_Bef_Irq = ch1; */
#define IRQ_When_Char_Ready 0x1
#define IRQ_When_THR_Empty 0x2

struct IrqConf
{
	void (*IrqActions)(void);
	unsigned char IR_Byte;
	unsigned char Chars_Bef_Irq;
};

typedef enum BAUD_RATE
{
	B110 = 0x0417,
	B150 = 0x0300,
	B300 = 0x0180,
	B600 = 0x00C0,
	B1200 = 0x0060,
	B2400 = 0x0030,
	B4800 = 0x0018,
	B9600 = 0x000C
} BAUD_RATE;

typedef enum SER_PORT
{
	SER_P1 = 0,
	SER_P2 = 2
} SER_PORT;

/* ====================================================*/
/* ==========   USEFUL MACROS        ==================*/
/* ====================================================*/

#define getoffset(a)	((unsigned)(0x0000FFFF & (unsigned long)(a)))
#define getsegment(a)	((unsigned)((0xFFFF0000 & (unsigned long)(a)) >> 16))
#define ptl(a)		(unsigned long) ( getoffset(a) +  (0x000FFFFF & (((unsigned long)getsegment(a))<<4) ) )
#define ltp(a)		(unsigned long) ( ((ulong)(a) % 16) + (((unsigned long)(a) / 16) << 16) )
#define	 setvec(_iv,_a)		*((char **)((unsigned)_iv*4)) = (char *)_a
#define  getvec(_iv)		(*((char **)((unsigned)_iv*4)))

/* ====================================================*/
/* ==========   FUNCTION DEFINITIONS   ================*/
/* ====================================================*/

/* Video Buffer screen functions (Can be used inside interrupt handlers) */
KERNEL_VOID PrintChar(int x, int y, char content);
KERNEL_VOID PrintString(int x, int y, char *str,int length);
/* Print 0-terminated string */
KERNEL_VOID PrintStr(int x, int y, char *string);
KERNEL_VOID PrintInteger(int x, int y, int number);
KERNEL_VOID cls(void);

/* Bios interrupt screen functions */
KERNEL_VOID bSetColor(SCR_COLORS fgcolor, SCR_COLORS bgcolor, unsigned int HighIntensity);
KERNEL_CHAR bGetColor();
KERNEL_VOID bSetCursor(int x, int y);
KERNEL_VOID bGetCursor(int *x, int *y);
KERNEL_VOID bcls(void);
KERNEL_VOID bAdvanceCursor(void);
KERNEL_VOID bBacktrackCursor(void);
KERNEL_VOID bNewLine(void);
KERNEL_VOID bPrintChar(char c);
KERNEL_VOID bPrintString(char *str,int length);
/* Print 0-terminated string */
KERNEL_VOID bPrintStr(char *str);
KERNEL_VOID bPrintInteger(int x);

/* Bios interrupt keyboard functions */
KERNEL_CHAR bGetChar(unsigned char *ExtendedCode);

/**************************************************************************************/
/*                                                                                    */
/*                   Serial Port Functions                                            */
/* There are 2 sets of functions. Use any of them but do not mix function calls       */
/*                      from different sets                                           */
/**************************************************************************************/

/**************************************************************************************/
/*                                                                                    */
/*                              1st Set                                               */
/*                              Original Functions                                    */
/*                                                                                    */
/**************************************************************************************/

/* Serial port functions (ORIGINAL)*/
/* Set HWDFlowControl to anything besides 0 to activate hardware flow control (NOT IMPLEMENTED YET) */
KERNEL_VOID SER_InitializePort(SER_PORT portnum, BAUD_RATE brate, const unsigned char HWDFlowControl, struct IrqConf *actions);
KERNEL_INT SER_ReadChar(SER_PORT portnum, char *ReadCh);
KERNEL_INT SER_WriteChar(SER_PORT portnum, char WriteCh);
KERNEL_VOID RestoreInterrupts(void);
/* Returns the cause of the interrupt (used only inside interrupt handlers) */
/* Returns 0 if the specified port hasn't triggered the interrupt */
/* Returns 1 if new char is available */
/* Returns 2 if a new char can be sent (THR Empty) */
/* Returns 3 if line status has changed */
/* Returns -1 if a transmission error occured */
KERNEL_INT InterruptReason(SER_PORT port);

/*************************************************************************************/
/*                                                                                   */
/*                              2nd Set                                              */
/*                              Simplified Functions (By default interrupt driven)   */
/*                                                                                   */
/*************************************************************************************/

/* Error_Type: */
/* 0 No error */
/* 1 Read error */
/* 3 Read driver internal buffer full. Char discarded */
/* 2 Write error */
struct QSR_ERROR
{
	unsigned char Error_Type;
	int Error;
};

KERNEL_VOID QSR_InitializePort(SER_PORT portnum, BAUD_RATE brate);

KERNEL_CHAR QSR_ReadChar(SER_PORT portnum);
KERNEL_INT QSR_WriteChar(SER_PORT portnum, char ch);

KERNEL_VOID QSR_FinalizePorts(void);

/* SEMAPHORE FUNCTIONS */
KERNEL_VOID P(int s);
KERNEL_VOID V(int s);
